package gs.test.v0

import cats.data.EitherT
import cats.Show
import gs.test.v0.GsTestError.TestDefinitionError.InvalidIterations

/**
  * Each instance of this class indicates the _definition_ of some test.
  *
  * @param name The display name of the test. Not considered to be unique.
  * @param permanentId The [[PermanentId]] for this test.
  * @param tags The set of [[Test.Tag]] applicable to this test.
  * @param iterations The number of iterations of this test to run.
  * @param f The effect that the test evaluates.
  */
final class TestDefinition[F[_]](
  val name: TestDefinition.Name,
  val permanentId: PermanentId,
  val documentation: Option[String],
  val tags: List[TestDefinition.Tag],
  val markers: List[TestDefinition.Marker],
  val iterations: TestDefinition.Iterations,
  val unitOfWork: EitherT[F, TestFailure, Unit]
)

object TestDefinition:

  /**
   * Opaque type representing names that may be assigned to [[Test]].
   */
  opaque type Name = String

  object Name:

    /**
     * Instantiate a new [[Test.Name]]. This name is not unique, has no 
     * constraints, and only exists for display purposes.
     *
     * @param name The candidate string.
     * @return The new [[Test.Name]] instance.
     */
    def apply(name: String): Name = name

    given CanEqual[Name, Name] = CanEqual.derived

    given Show[Name] = name => name

  end Name

  /**
   * Opaque type representing tags that may be assigned to a [[Test]].
   */
  opaque type Tag = String

  object Tag:

    /**
     * Instantiate a new [[Test.Tag]].
     *
     * @param tag The candidate string.
     * @return The new [[Test.Tag]] instance.
     */
    def apply(tag: String): Tag = tag

    given CanEqual[Tag, Tag] = CanEqual.derived

    given Show[Tag] = tag => tag

  end Tag

  /**
   * Enumeration for _Markers_, special tokens which "mark" a test to change
   * execution functionality.
   *
   * The basic case for this enumeration is allowing tests to be ignored.
   *
   * @param name The formal serialized name of the marker.
   */
  sealed abstract class Marker(val name: String)

  object Marker:

    given CanEqual[Marker, Marker] = CanEqual.derived

    /**
     * If this [[Test.Marker]] is present on a test, the test will be ignored.
     */
    case object Ignored extends Marker("ignored")

  end Marker

  /**
   * Opaque type that represents the number of iterations a test should run.
   * This value must be at least `1` (the default). To ignore a test, use the
   * [[Test.Marker.Ignored]] marker.
   */
  opaque type Iterations = Int

  object Iterations:

    def One: Iterations = 1

    /**
     * Validate and instantiate a new [[Iterations]] instance.
     *
     * @param candidate The candidate value. Must be 1 or greater.
     * @return The new [[Iterations]], or an error if an invalid input is given.
     */
    def validate(candidate: Int): Either[GsTestError, Iterations] =
      if candidate < 1 then 
        Left(InvalidIterations(candidate)) 
      else 
        Right(candidate)

    given CanEqual[Iterations, Iterations] = CanEqual.derived

    given Show[Iterations] = iters => iters.toString()

    extension (iters: Iterations)
      def toInt(): Int = iters

  end Iterations

end TestDefinition
